@@ -560,6 +560,11 @@ static const MethodMapping projectionMethodMappings[] = {
      EPSG_CODE_METHOD_LAMBERT_CONIC_CONFORMAL_2SP,
      "Lambert_Conformal_Conic_2SP", "lcc", nullptr, paramsLCC2SP},
 
+    // Oracle WKT
+    {EPSG_NAME_METHOD_LAMBERT_CONIC_CONFORMAL_2SP,
+     EPSG_CODE_METHOD_LAMBERT_CONIC_CONFORMAL_2SP, "Lambert Conformal Conic",
+     "lcc", nullptr, paramsLCC2SP},
+
     {EPSG_NAME_METHOD_LAMBERT_CONIC_CONFORMAL_2SP_MICHIGAN,
      EPSG_CODE_METHOD_LAMBERT_CONIC_CONFORMAL_2SP_MICHIGAN,
      nullptr, // no mapping to WKT1_GDAL
