@@ -18,20 +18,25 @@
 #include "proj.h"
 #include "proj_internal.h"
 #include <math.h>
-
+#include "mlfn.hpp"
 
 PROJ_HEAD(tmerc, "Transverse Mercator") "\n\tCyl, Sph&Ell\n\tapprox";
 PROJ_HEAD(etmerc, "Extended Transverse Mercator") "\n\tCyl, Sph";
 PROJ_HEAD(utm, "Universal Transverse Mercator (UTM)") "\n\tCyl, Ell\n\tzone= south approx";
 
 namespace { // anonymous namespace
-struct pj_opaque_approx {
+
+// Approximate: Evenden/Snyder
+struct EvendenSnyder
+{
     double  esp;
     double  ml0;
     double  *en;
 };
 
-struct pj_opaque_exact {
+// More exact: Poder/Engsager
+struct PoderEngsager
+{
     double    Qn;     /* Merid. quad., scaled to the projection */
     double    Zb;     /* Radius vector in polar coord. systems  */
     double    cgb[6]; /* Constants for Gauss -> Geo lat */
@@ -40,6 +45,11 @@ struct pj_opaque_exact {
     double    gtu[6]; /* Constants for geo -> transv. merc. */
 };
 
+struct tmerc_data {
+    EvendenSnyder approx;
+    PoderEngsager exact;
+};
+
 } // anonymous namespace
 
 /* Constants for "approximate" transverse mercator */
@@ -62,9 +72,11 @@ struct pj_opaque_exact {
 //
 /*****************************************************************************/
 
-static PJ_XY approx_e_fwd (PJ_LP lp, PJ *P) {
+
+static PJ_XY approx_e_fwd (PJ_LP lp, PJ *P)
+{
     PJ_XY xy = {0.0, 0.0};
-    struct pj_opaque_approx *Q = static_cast<struct pj_opaque_approx*>(P->opaque);
+    const auto *Q = &(static_cast<struct tmerc_data*>(P->opaque)->approx);
     double al, als, n, cosphi, sinphi, t;
 
     /*
@@ -94,7 +106,7 @@ static PJ_XY approx_e_fwd (PJ_LP lp, PJ *P) {
         FC5 * als * (5. + t * (t - 18.) + n * (14. - 58. * t)
         + FC7 * als * (61. + t * ( t * (179. - t) - 479. ) )
         )));
-    xy.y = P->k0 * (pj_mlfn(lp.phi, sinphi, cosphi, Q->en) - Q->ml0 +
+    xy.y = P->k0 * (inline_pj_mlfn(lp.phi, sinphi, cosphi, Q->en) - Q->ml0 +
         sinphi * al * lp.lam * FC2 * ( 1. +
         FC4 * als * (5. - t + n * (9. + 4. * n) +
         FC6 * als * (61. + t * (t - 58.) + n * (270. - 330 * t)
@@ -103,10 +115,10 @@ static PJ_XY approx_e_fwd (PJ_LP lp, PJ *P) {
     return (xy);
 }
 
-
 static PJ_XY approx_s_fwd (PJ_LP lp, PJ *P) {
     PJ_XY xy = {0.0,0.0};
     double b, cosphi;
+    const auto *Q = &(static_cast<struct tmerc_data*>(P->opaque)->approx);
 
     /*
      * Fail if our longitude is more than 90 degrees from the
@@ -129,7 +141,7 @@ static PJ_XY approx_s_fwd (PJ_LP lp, PJ *P) {
         return xy;
     }
 
-    xy.x = static_cast<struct pj_opaque_approx*>(P->opaque)->ml0 * log ((1. + b) / (1. - b));
+    xy.x = Q->ml0 * log ((1. + b) / (1. - b));
     xy.y = cosphi * cos (lp.lam) / sqrt (1. - b * b);
 
     b = fabs ( xy.y );
@@ -144,22 +156,20 @@ static PJ_XY approx_s_fwd (PJ_LP lp, PJ *P) {
 
     if (lp.phi < 0.)
         xy.y = -xy.y;
-    xy.y = static_cast<struct pj_opaque_approx*>(P->opaque)->esp * (xy.y - P->phi0);
+    xy.y = Q->esp * (xy.y - P->phi0);
     return xy;
 }
 
-
 static PJ_LP approx_e_inv (PJ_XY xy, PJ *P) {
     PJ_LP lp = {0.0,0.0};
-    struct pj_opaque_approx *Q = static_cast<struct pj_opaque_approx*>(P->opaque);
+    const auto *Q = &(static_cast<struct tmerc_data*>(P->opaque)->approx);
 
-    lp.phi = pj_inv_mlfn(P->ctx, Q->ml0 + xy.y / P->k0, P->es, Q->en);
+    double sinphi, cosphi;
+    lp.phi = inline_pj_inv_mlfn(P->ctx, Q->ml0 + xy.y / P->k0, P->es, Q->en, &sinphi, &cosphi);
     if (fabs(lp.phi) >= M_HALFPI) {
         lp.phi = xy.y < 0. ? -M_HALFPI : M_HALFPI;
         lp.lam = 0.;
     } else {
-        const double sinphi = sin(lp.phi);
-        const double cosphi = cos(lp.phi);
         double t = fabs (cosphi) > 1e-10 ? sinphi/cosphi : 0.;
         const double n = Q->esp * cosphi * cosphi;
         double con = 1. - P->es * sinphi * sinphi;
@@ -182,18 +192,18 @@ static PJ_LP approx_e_inv (PJ_XY xy, PJ *P) {
     return lp;
 }
 
-
 static PJ_LP approx_s_inv (PJ_XY xy, PJ *P) {
     PJ_LP lp = {0.0, 0.0};
     double h, g;
+    const auto *Q = &(static_cast<struct tmerc_data*>(P->opaque)->approx);
 
-    h = exp(xy.x / static_cast<struct pj_opaque_approx*>(P->opaque)->esp);
+    h = exp(xy.x / Q->esp);
     if( h == 0 ) {
         proj_errno_set(P, PJD_ERR_INVALID_X_OR_Y);
         return proj_coord_error().lp;
     }
     g = .5 * (h - 1. / h);
-    h = cos (P->phi0 + xy.y / static_cast<struct pj_opaque_approx*>(P->opaque)->esp);
+    h = cos (P->phi0 + xy.y / Q->esp);
     lp.phi = asin(sqrt((1. - h * h) / (1. + g * g)));
 
     /* Make sure that phi is on the correct hemisphere when false northing is used */
@@ -204,22 +214,20 @@ static PJ_LP approx_s_inv (PJ_XY xy, PJ *P) {
 }
 
 
-static PJ *destructor_approx(PJ *P, int errlev) {
+static PJ *destructor(PJ *P, int errlev) {
     if (nullptr==P)
         return nullptr;
 
     if (nullptr==P->opaque)
         return pj_default_destructor(P, errlev);
 
-    pj_dealloc (static_cast<struct pj_opaque_approx*>(P->opaque)->en);
+    pj_dealloc (static_cast<struct tmerc_data*>(P->opaque)->approx.en);
     return pj_default_destructor(P, errlev);
 }
 
 
 static PJ *setup_approx(PJ *P) {
-    struct pj_opaque_approx *Q = static_cast<struct pj_opaque_approx*>(P->opaque);
-
-    P->destructor = destructor_approx;
+    auto *Q = &(static_cast<struct tmerc_data*>(P->opaque)->approx);
 
     if (P->es != 0.0) {
         if (!(Q->en = pj_enfn(P->es)))
@@ -227,13 +235,9 @@ static PJ *setup_approx(PJ *P) {
 
         Q->ml0 = pj_mlfn(P->phi0, sin(P->phi0), cos(P->phi0), Q->en);
         Q->esp = P->es / (1. - P->es);
-        P->inv = approx_e_inv;
-        P->fwd = approx_e_fwd;
     } else {
         Q->esp = P->k0;
         Q->ml0 = .5 * Q->esp;
-        P->inv = approx_s_inv;
-        P->fwd = approx_s_fwd;
     }
     return P;
 }
@@ -258,43 +262,32 @@ static PJ *setup_approx(PJ *P) {
 //
 /*****************************************************************************/
 
-/* Helper functios for "exact" transverse mercator */
-#ifdef _GNU_SOURCE
-    inline
-#endif
-static double gatg(double *p1, int len_p1, double B) {
-    double *p;
-    double h = 0, h1, h2 = 0, cos_2B;
+/* Helper functions for "exact" transverse mercator */
+inline
+static double gatg(const double *p1, int len_p1, double B, double cos_2B, double sin_2B) {
+    double h = 0, h1, h2 = 0;
 
-    cos_2B = 2*cos(2*B);
-    p = p1 + len_p1;
+    const double two_cos_2B = 2*cos_2B;
+    const double* p = p1 + len_p1;
     h1 = *--p;
     while (p - p1) {
-        h = -h2 + cos_2B*h1 + *--p;
+        h = -h2 + two_cos_2B*h1 + *--p;
         h2 = h1;
         h1 = h;
     }
-    return (B + h*sin(2*B));
+    return (B + h*sin_2B);
 }
 
 /* Complex Clenshaw summation */
-#ifdef _GNU_SOURCE
-    inline
-#endif
-static double clenS(double *a, int size, double arg_r, double arg_i, double *R, double *I) {
-    double      *p, r, i, hr, hr1, hr2, hi, hi1, hi2;
-    double      sin_arg_r, cos_arg_r, sinh_arg_i, cosh_arg_i;
+inline
+static double clenS(const double *a, int size,
+                    double sin_arg_r, double cos_arg_r,
+                    double sinh_arg_i, double cosh_arg_i,
+                    double *R, double *I) {
+    double      r, i, hr, hr1, hr2, hi, hi1, hi2;
 
     /* arguments */
-    p = a + size;
-#ifdef _GNU_SOURCE
-    sincos(arg_r, &sin_arg_r, &cos_arg_r);
-#else
-    sin_arg_r  = sin(arg_r);
-    cos_arg_r  = cos(arg_r);
-#endif
-    sinh_arg_i = sinh(arg_i);
-    cosh_arg_i = cosh(arg_i);
+    const double* p = a + size;
     r          =  2*cos_arg_r*cosh_arg_i;
     i          = -2*sin_arg_r*sinh_arg_i;
 
@@ -319,10 +312,10 @@ static double clenS(double *a, int size, double arg_r, double arg_i, double *R,
 
 
 /* Real Clenshaw summation */
-static double clens(double *a, int size, double arg_r) {
-    double      *p, r, hr, hr1, hr2, cos_arg_r;
+static double clens(const double *a, int size, double arg_r) {
+    double      r, hr, hr1, hr2, cos_arg_r;
 
-    p = a + size;
+    const double* p = a + size;
     cos_arg_r  = cos(arg_r);
     r          =  2*cos_arg_r;
 
@@ -340,29 +333,78 @@ static double clens(double *a, int size, double arg_r) {
 /* Ellipsoidal, forward */
 static PJ_XY exact_e_fwd (PJ_LP lp, PJ *P) {
     PJ_XY xy = {0.0,0.0};
-    struct pj_opaque_exact *Q = static_cast<struct pj_opaque_exact*>(P->opaque);
-    double sin_Cn, cos_Cn, cos_Ce, sin_Ce, dCn, dCe;
-    double Cn = lp.phi, Ce = lp.lam;
+    const auto *Q = &(static_cast<struct tmerc_data*>(P->opaque)->exact);
 
     /* ell. LAT, LNG -> Gaussian LAT, LNG */
-    Cn  = gatg (Q->cbg, PROJ_ETMERC_ORDER, Cn);
+    double Cn  = gatg (Q->cbg, PROJ_ETMERC_ORDER, lp.phi, cos(2*lp.phi), sin(2*lp.phi));
     /* Gaussian LAT, LNG -> compl. sph. LAT */
-#ifdef _GNU_SOURCE
-    sincos (Cn, &sin_Cn, &cos_Cn);
-    sincos (Ce, &sin_Ce, &cos_Ce);
-#else
-    sin_Cn = sin (Cn);
-    cos_Cn = cos (Cn);
-    sin_Ce = sin (Ce);
-    cos_Ce = cos (Ce);
+    const double sin_Cn = sin (Cn);
+    const double cos_Cn = cos (Cn);
+    const double sin_Ce = sin (lp.lam);
+    const double cos_Ce = cos (lp.lam);
+
+    const double cos_Cn_cos_Ce = cos_Cn*cos_Ce;
+    Cn     = atan2 (sin_Cn, cos_Cn_cos_Ce);
+
+    const double inv_denom_tan_Ce = 1. / hypot (sin_Cn, cos_Cn_cos_Ce);
+    const double tan_Ce = sin_Ce*cos_Cn * inv_denom_tan_Ce;
+#if 0
+    // Variant of the above: found not to be measurably faster
+    const double sin_Ce_cos_Cn = sin_Ce*cos_Cn;
+    const double denom = sqrt(1 - sin_Ce_cos_Cn * sin_Ce_cos_Cn);
+    const double tan_Ce = sin_Ce_cos_Cn / denom;
 #endif
 
-    Cn     = atan2 (sin_Cn, cos_Ce*cos_Cn);
-    Ce     = atan2 (sin_Ce*cos_Cn,  hypot (sin_Cn, cos_Cn*cos_Ce));
-
     /* compl. sph. N, E -> ell. norm. N, E */
-    Ce  = asinh ( tan (Ce) );     /* Replaces: Ce  = log(tan(FORTPI + Ce*0.5)); */
-    Cn += clenS (Q->gtu, PROJ_ETMERC_ORDER, 2*Cn, 2*Ce, &dCn, &dCe);
+    double Ce  = asinh ( tan_Ce );     /* Replaces: Ce  = log(tan(FORTPI + Ce*0.5)); */
+
+/*
+ *  Non-optimized version:
+ *  const double sin_arg_r  = sin(2*Cn);
+ *  const double cos_arg_r  = cos(2*Cn);
+ *
+ *  Given:
+ *      sin(2 * Cn) = 2 sin(Cn) cos(Cn)
+ *          sin(atan(y)) = y / sqrt(1 + y^2)
+ *          cos(atan(y)) = 1 / sqrt(1 + y^2)
+ *      ==> sin(2 * Cn) = 2 tan_Cn / (1 + tan_Cn^2)
+ *
+ *      cos(2 * Cn) = 2cos^2(Cn) - 1
+ *                  = 2 / (1 + tan_Cn^2) - 1
+ */
+    const double two_inv_denom_tan_Ce = 2 * inv_denom_tan_Ce;
+    const double two_inv_denom_tan_Ce_square = two_inv_denom_tan_Ce * inv_denom_tan_Ce;
+    const double tmp_r = cos_Cn_cos_Ce * two_inv_denom_tan_Ce_square;
+    const double sin_arg_r  = sin_Cn * tmp_r;
+    const double cos_arg_r  = cos_Cn_cos_Ce * tmp_r - 1;
+
+/*
+ *  Non-optimized version:
+ *  const double sinh_arg_i = sinh(2*Ce);
+ *  const double cosh_arg_i = cosh(2*Ce);
+ *
+ *  Given
+ *      sinh(2 * Ce) = 2 sinh(Ce) cosh(Ce)
+ *          sinh(asinh(y)) = y
+ *          cosh(asinh(y)) = sqrt(1 + y^2)
+ *      ==> sinh(2 * Ce) = 2 tan_Ce sqrt(1 + tan_Ce^2)
+ *
+ *      cosh(2 * Ce) = 2cosh^2(Ce) - 1
+ *                   = 2 * (1 + tan_Ce^2) - 1
+ *
+ * and 1+tan_Ce^2 = 1 + sin_Ce^2 * cos_Cn^2 / (sin_Cn^2 + cos_Cn^2 * cos_Ce^2)
+ *                = (sin_Cn^2 + cos_Cn^2 * cos_Ce^2 + sin_Ce^2 * cos_Cn^2) / (sin_Cn^2 + cos_Cn^2 * cos_Ce^2)
+ *                = 1. / (sin_Cn^2 + cos_Cn^2 * cos_Ce^2)
+ *                = inv_denom_tan_Ce^2
+ *
+ */
+    const double sinh_arg_i = tan_Ce * two_inv_denom_tan_Ce;
+    const double cosh_arg_i = two_inv_denom_tan_Ce_square - 1;
+
+    double dCn, dCe;
+    Cn += clenS (Q->gtu, PROJ_ETMERC_ORDER,
+                 sin_arg_r, cos_arg_r, sinh_arg_i, cosh_arg_i,
+                 &dCn, &dCe);
     Ce += dCe;
     if (fabs (Ce) <= 2.623395162778) {
         xy.y  = Q->Qn * Cn + Q->Zb;  /* Northing */
@@ -376,33 +418,69 @@ static PJ_XY exact_e_fwd (PJ_LP lp, PJ *P) {
 /* Ellipsoidal, inverse */
 static PJ_LP exact_e_inv (PJ_XY xy, PJ *P) {
     PJ_LP lp = {0.0,0.0};
-    struct pj_opaque_exact *Q = static_cast<struct pj_opaque_exact*>(P->opaque);
-    double sin_Cn, cos_Cn, cos_Ce, sin_Ce, dCn, dCe;
-    double Cn = xy.y, Ce = xy.x;
+    const auto *Q = &(static_cast<struct tmerc_data*>(P->opaque)->exact);
 
     /* normalize N, E */
-    Cn = (Cn - Q->Zb)/Q->Qn;
-    Ce = Ce/Q->Qn;
+    double Cn = (xy.y - Q->Zb)/Q->Qn;
+    double Ce = xy.x/Q->Qn;
 
     if (fabs(Ce) <= 2.623395162778) { /* 150 degrees */
         /* norm. N, E -> compl. sph. LAT, LNG */
-        Cn += clenS(Q->utg, PROJ_ETMERC_ORDER, 2*Cn, 2*Ce, &dCn, &dCe);
+        const double sin_arg_r  = sin(2*Cn);
+        const double cos_arg_r  = cos(2*Cn);
+
+        //const double sinh_arg_i = sinh(2*Ce);
+        //const double cosh_arg_i = cosh(2*Ce);
+        const double exp_2_Ce = exp(2*Ce);
+        const double half_inv_exp_2_Ce = 0.5 / exp_2_Ce;
+        const double sinh_arg_i = 0.5 * exp_2_Ce - half_inv_exp_2_Ce;
+        const double cosh_arg_i = 0.5 * exp_2_Ce + half_inv_exp_2_Ce;
+
+        double dCn_ignored, dCe;
+        Cn += clenS(Q->utg, PROJ_ETMERC_ORDER,
+                    sin_arg_r, cos_arg_r, sinh_arg_i, cosh_arg_i,
+                    &dCn_ignored, &dCe);
         Ce += dCe;
-        Ce = atan (sinh (Ce)); /* Replaces: Ce = 2*(atan(exp(Ce)) - FORTPI); */
+
         /* compl. sph. LAT -> Gaussian LAT, LNG */
-#ifdef _GNU_SOURCE
-        sincos (Cn, &sin_Cn, &cos_Cn);
-        sincos (Ce, &sin_Ce, &cos_Ce);
-#else
-        sin_Cn = sin (Cn);
-        cos_Cn = cos (Cn);
+        const double sin_Cn = sin (Cn);
+        const double cos_Cn = cos (Cn);
+
+#if 0
+        // Non-optimized version:
+        double sin_Ce, cos_Ce;
+        Ce = atan (sinh (Ce));  // Replaces: Ce = 2*(atan(exp(Ce)) - FORTPI);
         sin_Ce = sin (Ce);
         cos_Ce = cos (Ce);
-#endif
         Ce     = atan2 (sin_Ce, cos_Ce*cos_Cn);
         Cn     = atan2 (sin_Cn*cos_Ce,  hypot (sin_Ce, cos_Ce*cos_Cn));
+#else
+/*
+ *      One can divide both member of Ce = atan2(...) by cos_Ce, which gives:
+ *      Ce     = atan2 (tan_Ce, cos_Cn) = atan2(sinh(Ce), cos_Cn)
+ *
+ *      and the same for Cn = atan2(...)
+ *      Cn     = atan2 (sin_Cn, hypot (sin_Ce, cos_Ce*cos_Cn)/cos_Ce)
+ *             = atan2 (sin_Cn, hypot (sin_Ce/cos_Ce, cos_Cn))
+ *             = atan2 (sin_Cn, hypot (tan_Ce, cos_Cn))
+ *             = atan2 (sin_Cn, hypot (sinhCe, cos_Cn))
+ */
+        const double sinhCe = sinh (Ce);
+        Ce     = atan2 (sinhCe, cos_Cn);
+        const double modulus_Ce = hypot (sinhCe, cos_Cn);
+        Cn     = atan2 (sin_Cn, modulus_Ce);
+#endif
+
         /* Gaussian LAT, LNG -> ell. LAT, LNG */
-        lp.phi = gatg (Q->cgb,  PROJ_ETMERC_ORDER, Cn);
+
+        // Optimization of the computation of cos(2*Cn) and sin(2*Cn)
+        const double tmp = 2 * modulus_Ce / (sinhCe * sinhCe + 1);
+        const double sin_2_Cn  = sin_Cn * tmp;
+        const double cos_2_Cn  = tmp * modulus_Ce - 1.;
+        //const double cos_2_Cn = cos(2 * Cn);
+        //const double sin_2_Cn = sin(2 * Cn);
+
+        lp.phi = gatg (Q->cgb,  PROJ_ETMERC_ORDER, Cn, cos_2_Cn, sin_2_Cn);
         lp.lam = Ce;
     }
     else
@@ -412,11 +490,9 @@ static PJ_LP exact_e_inv (PJ_XY xy, PJ *P) {
 
 static PJ *setup_exact(PJ *P) {
     double f, n, np, Z;
-    struct pj_opaque_exact *Q = static_cast<struct pj_opaque_exact*>(P->opaque);
+    auto *Q = &(static_cast<struct tmerc_data*>(P->opaque)->exact);
 
-    if (P->es <= 0) {
-        return pj_default_destructor(P, PJD_ERR_ELLIPSOID_USE_REQUIRED);
-    }
+    assert( P->es > 0 );
 
     /* flattening */
     f = P->es / (1 + sqrt (1 -  P->es)); /* Replaces: f = 1 - sqrt(1-P->es); */
@@ -487,18 +563,141 @@ static PJ *setup_exact(PJ *P) {
     Q->gtu[5] = np*(212378941/319334400.0);
 
     /* Gaussian latitude value of the origin latitude */
-    Z = gatg (Q->cbg, PROJ_ETMERC_ORDER, P->phi0);
+    Z = gatg (Q->cbg, PROJ_ETMERC_ORDER, P->phi0, cos(2*P->phi0), sin(2*P->phi0));
 
     /* Origin northing minus true northing at the origin latitude */
     /* i.e. true northing = N - P->Zb                         */
     Q->Zb  = - Q->Qn*(Z + clens(Q->gtu, PROJ_ETMERC_ORDER, 2*Z));
-    P->inv = exact_e_inv;
-    P->fwd = exact_e_fwd;
+
     return P;
 }
 
 
 
+static PJ_XY auto_e_fwd (PJ_LP lp, PJ *P) {
+    if( fabs(lp.lam) > 3 * DEG_TO_RAD )
+        return exact_e_fwd(lp, P);
+    else
+        return approx_e_fwd(lp, P);
+}
+
+static PJ_LP auto_e_inv (PJ_XY xy, PJ *P) {
+    // For k = 1 and lon = 3 (from central meridian),
+    // At lat = 0, we get x ~= 0.052, y = 0
+    // At lat = 90, we get x = 0, y ~= 1.57
+    // And the shape of this x=f(y) frontier curve is very very roughly a
+    // parabola. Hence:
+    if( fabs(xy.x) > 0.053 - 0.022 * xy.y * xy.y )
+        return exact_e_inv(xy, P);
+    else
+        return approx_e_inv(xy, P);
+}
+
+static PJ *setup(PJ *P, TMercAlgo eAlg) {
+
+    struct tmerc_data *Q = static_cast<struct tmerc_data*>(pj_calloc (1, sizeof (struct tmerc_data)));
+    if (nullptr==Q)
+        return pj_default_destructor (P, ENOMEM);
+    P->opaque = Q;
+
+    if( P->es == 0 )
+        eAlg = TMercAlgo::EVENDEN_SNYDER;
+
+    switch( eAlg )
+    {
+        case TMercAlgo::EVENDEN_SNYDER:
+        {
+            P->destructor = destructor;
+            if( !setup_approx(P) )
+                return nullptr;
+            if( P->es == 0 )
+            {
+                P->inv = approx_s_inv;
+                P->fwd = approx_s_fwd;
+            }
+            else
+            {
+                P->inv = approx_e_inv;
+                P->fwd = approx_e_fwd;
+            }
+            break;
+        }
+
+        case TMercAlgo::PODER_ENGSAGER:
+        {
+            setup_exact(P);
+            P->inv = exact_e_inv;
+            P->fwd = exact_e_fwd;
+            break;
+        }
+
+        case TMercAlgo::AUTO:
+        {
+            P->destructor = destructor;
+            if( !setup_approx(P) )
+                return nullptr;
+            setup_exact(P);
+
+            P->inv = auto_e_inv;
+            P->fwd = auto_e_fwd;
+            break;
+        }
+    }
+    return P;
+}
+
+
+static bool getAlgoFromParams(PJ* P, TMercAlgo& algo)
+{
+    if( pj_param (P->ctx, P->params, "bapprox").i )
+    {
+        algo = TMercAlgo::EVENDEN_SNYDER;
+        return true;
+    }
+
+    const char* algStr = pj_param (P->ctx, P->params, "salgo").s;
+    if( algStr )
+    {
+        if( strcmp(algStr, "evenden_snyder") == 0 )
+        {
+            algo = TMercAlgo::EVENDEN_SNYDER;
+            return true;
+        }
+        if( strcmp(algStr, "poder_engsager") == 0 )
+        {
+            algo = TMercAlgo::PODER_ENGSAGER;
+            return true;
+        }
+        if( strcmp(algStr, "auto") == 0 )
+        {
+            algo = TMercAlgo::AUTO;
+            // Don't return so that we can run a later validity check
+        }
+        else
+        {
+            proj_log_error (P, "unknown value for +algo");
+            return false;
+        }
+    }
+    else
+    {
+        pj_load_ini(P->ctx); // if not already done
+        pj_ctx_set_errno(P->ctx, 0); // reset error in case proj.ini couldn't be found
+        algo = P->ctx->defaultTmercAlgo;
+    }
+
+    // We haven't worked on the criterion on inverse transformation
+    // when phi0 != 0 or if k0 is not close to 1 or for very oblate
+    // ellipsoid (es > 0.1 is ~ rf < 200)
+    if( algo == TMercAlgo::AUTO &&
+            (P->es > 0.1 || P->phi0 != 0 || fabs(P->k0 - 1) > 0.01) )
+    {
+        algo = TMercAlgo::PODER_ENGSAGER;
+    }
+
+    return true;
+}
+
 
 /*****************************************************************************/
 //
@@ -509,35 +708,25 @@ static PJ *setup_exact(PJ *P) {
 PJ *PROJECTION(tmerc) {
     /* exact transverse mercator only exists in ellipsoidal form, */
     /* use approximate version if +a sphere is requested          */
-    if (pj_param (P->ctx, P->params, "bapprox").i || P->es <= 0) {
-        struct pj_opaque_approx *Q = static_cast<struct pj_opaque_approx*>(pj_calloc (1, sizeof (struct pj_opaque_approx)));
-        if (nullptr==Q)
-            return pj_default_destructor (P, ENOMEM);
 
-        P->opaque = Q;
-
-        return setup_approx(P);
-    } else {
-        struct pj_opaque_exact *Q = static_cast<struct pj_opaque_exact*>(pj_calloc (1, sizeof (struct pj_opaque_exact)));
-        if (nullptr==Q)
-            return pj_default_destructor (P, ENOMEM);
-        P->opaque = Q;
-       return setup_exact (P);
-    }
+    TMercAlgo algo;
+    if( !getAlgoFromParams(P, algo) )
+        return pj_default_destructor(P, PJD_ERR_INVALID_ARG);
+    return setup(P, algo);
 }
 
 
 PJ *PROJECTION(etmerc) {
-    struct pj_opaque_exact *Q = static_cast<struct pj_opaque_exact*>(pj_calloc (1, sizeof (struct pj_opaque_exact)));
-    if (nullptr==Q)
-        return pj_default_destructor (P, ENOMEM);
-    P->opaque = Q;
-   return setup_exact (P);
+    if (P->es == 0.0) {
+        return pj_default_destructor(P, PJD_ERR_ELLIPSOID_USE_REQUIRED);
+    }
+
+   return setup (P, TMercAlgo::PODER_ENGSAGER);
 }
 
 
 /* UTM uses the Poder/Engsager implementation for the underlying projection      */
-/* UNLESS +approx is set in which case the Evenden/Snyder implemenation is used. */
+/* UNLESS +approx is set in which case the Evenden/Snyder implementation is used. */
 PJ *PROJECTION(utm) {
     long zone;
     if (P->es == 0.0) {
@@ -570,19 +759,8 @@ PJ *PROJECTION(utm) {
     P->k0 = 0.9996;
     P->phi0 = 0.;
 
-    if (pj_param(P->ctx, P->params, "bapprox").i) {
-        struct pj_opaque_approx *Q = static_cast<struct pj_opaque_approx*>(pj_calloc (1, sizeof (struct pj_opaque_approx)));
-        if (nullptr==Q)
-            return pj_default_destructor (P, ENOMEM);
-        P->opaque = Q;
-
-        return setup_approx(P);
-    } else {
-        struct pj_opaque_exact *Q = static_cast<struct pj_opaque_exact*>(pj_calloc (1, sizeof (struct pj_opaque_exact)));
-        if (nullptr==Q)
-            return pj_default_destructor (P, ENOMEM);
-        P->opaque = Q;
-
-        return setup_exact(P);
-    }
+    TMercAlgo algo;
+    if( !getAlgoFromParams(P, algo) )
+        return pj_default_destructor(P, PJD_ERR_INVALID_ARG);
+    return setup(P, algo);
 }
